use anyhow::Context;
use ssh_key::{Certificate, PublicKey};
use std::{
    fmt::Debug,
    io::ErrorKind,
    path::{Path, PathBuf},
};
use tokio::fs;
use tracing::{instrument, trace};

#[derive(Debug, thiserror::Error)]
pub enum CertError {
    #[error("missing key id")]
    NoKID,
    #[error("missing ca identifier (comment)")]
    NoCAComment,
}

pub async fn read_certs(
    ca: &PublicKey,
    path: impl AsRef<Path>,
) -> anyhow::Result<Vec<Certificate>> {
    let ca_dir = path.as_ref().join(ca_dir(ca)?);
    if !ca_dir.exists() {
        return Ok(Vec::new());
    }
    read_certs_dir(&ca_dir).await
}

#[instrument]
pub async fn read_certs_dir(path: impl AsRef<Path> + Debug) -> anyhow::Result<Vec<Certificate>> {
    let mut dir = fs::read_dir(path.as_ref())
        .await
        .with_context(|| format!("read certs dir '{:?}'", path.as_ref()))?;
    let mut certs = Vec::new();
    while let Some(entry) = dir.next_entry().await? {
        //TODO: investigate why path().ends_with doesn't work
        if !entry
            .file_name()
            .into_string()
            .unwrap()
            .ends_with("-cert.pub")
        {
            trace!(
                "skipped {:?} due to missing '-cert.pub' extension",
                entry.path()
            );
            continue;
        }
        let cert = load_cert(entry.path()).await?;
        if let Some(cert) = cert {
            certs.push(cert);
        }
    }
    Ok(certs)
}

pub async fn read_pubkey_dir(path: impl AsRef<Path> + Debug) -> anyhow::Result<Vec<PublicKey>> {
    let mut dir = fs::read_dir(path.as_ref())
        .await
        .with_context(|| format!("read certs dir '{:?}'", path.as_ref()))?;
    let mut pubs = Vec::new();
    while let Some(entry) = dir.next_entry().await? {
        //TODO: investigate why path().ends_with doesn't work
        let file_name = entry.file_name().into_string().unwrap();
        if !file_name.ends_with(".pub") || file_name.ends_with("-cert.pub")
        {
            trace!(
                "skipped {:?} due to missing '.pub' extension",
                entry.path()
            );
            continue;
        }
        let cert = load_public_key(entry.path()).await?;
        if let Some(cert) = cert {
            pubs.push(cert);
        }
    }
    Ok(pubs)

}

fn parse_utf8(bytes: Vec<u8>) -> anyhow::Result<String> {
    String::from_utf8(bytes).context("invalid utf-8")
}

pub async fn read_pubkey(path: impl AsRef<Path>) -> anyhow::Result<PublicKey> {
    let contents = fs::read(&path)
        .await
        .with_context(|| format!("read {:?}", path.as_ref()))?;
    let string_repr = parse_utf8(contents)?;
    PublicKey::from_openssh(&string_repr)
        .with_context(|| format!("parse '{}' as public key", string_repr))
}

fn ca_dir(ca: &PublicKey) -> Result<String, CertError> {
    if ca.comment().is_empty() {
        return Err(CertError::NoCAComment);
    }
    Ok(ca.comment().to_string())
}

#[instrument]
fn cert_path(ca: &PublicKey, identifier: &str) -> Result<String, CertError> {
    if identifier.is_empty() {
        return Err(CertError::NoKID);
    }
    Ok(format!("{}/{}-cert.pub", ca_dir(ca)?, identifier))
}

#[instrument]
pub async fn store_cert(
    cert_dir: impl AsRef<Path> + Debug,
    ca: &PublicKey,
    cert: &Certificate,
) -> anyhow::Result<PathBuf> {
    // TODO: proper store
    let path = cert_dir.as_ref().join(cert_path(&ca, cert.key_id())?);
    if let Some(parent) = path.parent() {
        fs::create_dir_all(parent)
            .await
            .with_context(|| format!("mkdir -p {parent:?}"))?;
    }
    fs::write(&path, cert.to_openssh().context("encode cert")?)
        .await
        .context("write cert")?;
    Ok(path)
}

pub async fn load_cert_by_id(
    cert_dir: impl AsRef<Path>,
    ca: &PublicKey,
    identifier: &str,
) -> anyhow::Result<Option<Certificate>> {
    let path = cert_dir.as_ref().join(cert_path(ca, identifier)?);
    load_cert(&path).await
}

pub async fn load_cert(file: impl AsRef<Path> + Debug) -> anyhow::Result<Option<Certificate>> {
    let contents = match fs::read(&file).await {
        Ok(contents) => contents,
        Err(e) if e.kind() == ErrorKind::NotFound => return Ok(None),
        Err(e) => return Err(e).with_context(|| format!("read {:?}", &file)),
    };
    let string_repr = parse_utf8(contents)?;
    Ok(Some(Certificate::from_openssh(&string_repr).with_context(
        || format!("parse {:?} as openssh certificate", &file),
    )?))
}

pub async fn load_public_key(file: impl AsRef<Path> + Debug) -> anyhow::Result<Option<PublicKey>> {
    let contents = match fs::read(&file).await {
        Ok(contents) => contents,
        Err(e) if e.kind() == ErrorKind::NotFound => return Ok(None),
        Err(e) => return Err(e).with_context(|| format!("read {:?}", &file)),
    };
    let string_repr = parse_utf8(contents)?;
    Ok(Some(PublicKey::from_openssh(&string_repr).with_context(
        || format!("parse {:?} as openssh public key", &file),
    )?))

}
